import { ComputeBudgetProgram, Transaction, } from "@solana/web3.js";
import { chunks } from "./accountFetchCache";
import bs58 from "bs58";
import { ProgramError } from "./anchorError";
async function sleep(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
async function promiseAllInOrder(it) {
    let ret = [];
    for (const i of it) {
        ret.push(await i());
    }
    return ret;
}
export async function sendInstructions(provider, instructions, signers = [], payer = provider.wallet.publicKey, commitment = "confirmed", idlErrors = new Map()) {
    if (instructions.length == 0) {
        return "";
    }
    let tx = new Transaction();
    tx.recentBlockhash = (await provider.connection.getLatestBlockhash()).blockhash;
    tx.feePayer = payer || provider.wallet.publicKey;
    tx.add(...instructions);
    if (signers.length > 0) {
        tx.partialSign(...signers);
    }
    if (tx.feePayer.equals(provider.wallet.publicKey) ||
        tx.instructions.some((ix) => ix.keys.some((key) => key.isSigner && key.pubkey.equals(provider.wallet.publicKey)))) {
        tx = await provider.wallet.signTransaction(tx);
    }
    try {
        const { txid } = await sendAndConfirmWithRetry(provider.connection, tx.serialize(), {
            skipPreflight: true,
        }, commitment);
        return txid;
    }
    catch (e) {
        console.error(e);
        if (e.logs) {
            console.error(`Logs: ${e.logs}`);
        }
        const wrappedE = ProgramError.parse(e, idlErrors);
        throw wrappedE == null ? e : wrappedE;
    }
}
function truthy(value) {
    return !!value;
}
export async function sendMultipleInstructions(provider, instructionGroups, signerGroups, payer, finality = "confirmed", idlErrors = new Map()) {
    const recentBlockhash = (await provider.connection.getLatestBlockhash("confirmed")).blockhash;
    const ixAndSigners = instructionGroups
        .map((instructions, i) => {
        const signers = signerGroups[i];
        return {
            instructions,
            signers,
        };
    })
        .filter(({ instructions }) => instructions.length > 0);
    const txns = ixAndSigners.map(({ instructions }) => {
        const tx = new Transaction({
            feePayer: payer || provider.wallet.publicKey,
            recentBlockhash,
        });
        tx.add(...instructions);
        return tx;
    });
    const txnsSignedByWallet = await provider.wallet.signAllTransactions(txns);
    const txnsSigned = txnsSignedByWallet
        .map((tx, index) => {
        const signers = ixAndSigners[index].signers;
        if (signers.length > 0) {
            tx.partialSign(...signers);
        }
        return tx;
    })
        .map((tx) => tx.serialize());
    console.log("Sending multiple transactions...");
    try {
        return await promiseAllInOrder(txnsSigned.map((txn) => async () => {
            const { txid } = await sendAndConfirmWithRetry(provider.connection, txn, {
                skipPreflight: true,
            }, finality);
            return txid;
        }));
    }
    catch (e) {
        console.error(e);
        const wrappedE = ProgramError.parse(e, idlErrors);
        throw wrappedE == null ? e : wrappedE;
    }
}
export async function execute(program, provider, command, payer = provider.wallet.publicKey, commitment) {
    const { instructions, signers, output } = command;
    const errors = program.idl.errors?.reduce((acc, err) => {
        acc.set(err.code, `${err.name}: ${err.msg}`);
        return acc;
    }, new Map());
    if (instructions.length > 0) {
        const txid = await sendInstructions(provider, instructions, signers, payer, commitment, errors);
        return { txid, ...output };
    }
    // @ts-ignore
    return output;
}
export async function executeBig(program, provider, command, payer = provider.wallet.publicKey, finality) {
    const { instructions, signers, output } = command;
    const errors = program.idl.errors?.reduce((acc, err) => {
        acc.set(err.code, `${err.name}: ${err.msg}`);
        return acc;
    }, new Map());
    if (instructions.length > 0) {
        const txids = await sendMultipleInstructions(provider, instructions, signers, payer || provider.wallet.publicKey, finality, errors);
        return {
            ...output,
            txids: Array.from(txids),
        };
    }
    // @ts-ignore
    return output;
}
function getUnixTime() {
    return new Date().valueOf() / 1000;
}
const getPriorityFeeEstimate = async (connection, transaction, priorityLevel = 'NONE') => {
    const response = await fetch(connection.rpcEndpoint, {
        method: "POST",
        headers: { "Content-Type": "application/json" },
        body: JSON.stringify({
            jsonrpc: "2.0",
            id: "1",
            method: "getPriorityFeeEstimate",
            params: [
                {
                    transaction: bs58.encode(transaction.serialize()),
                    options: { priority_level: priorityLevel },
                },
            ],
        }),
    });
    const data = await response.json();
    console.log("Fee in function for", priorityLevel, " :", 
    // @ts-ignore
    data.result.priorityFeeEstimate);
    // @ts-ignore
    return data.result;
};
export const sendMultipleInstructionsWithPrioFee = (provider, instructionGroups, signerGroups, payer, priorityLevel = 'NONE', finality = "confirmed", timeout = DEFAULT_TIMEOUT, idlErrors = new Map()) => {
    const sendMultipleTxs = async (priorityLevel) => {
        const recentBlockhash = (await provider.connection.getLatestBlockhash("confirmed")).blockhash;
        const ixAndSigners = instructionGroups
            .map((instructions, i) => {
            const signers = signerGroups[i];
            return {
                instructions,
                signers,
            };
        })
            .filter(({ instructions }) => instructions.length > 0);
        const txns = ixAndSigners.map(({ instructions }) => {
            const tx = new Transaction({
                feePayer: payer || provider.wallet.publicKey,
                recentBlockhash,
            });
            tx.add(...instructions);
            return tx;
        });
        const txnsSignedByWallet = await provider.wallet.signAllTransactions(txns);
        let txnsSigned = txnsSignedByWallet
            .map((tx, index) => {
            const signers = ixAndSigners[index].signers;
            if (signers.length > 0) {
                tx.partialSign(...signers);
            }
            return tx;
        });
        if (priorityLevel !== "NONE") {
            const feeEstimates = await Promise.all(txnsSigned.map(tx => getPriorityFeeEstimate(provider.connection, tx, priorityLevel)));
            for (let i = 0; i < feeEstimates.length; i += 1) {
                const feeEstimate = feeEstimates[i];
                const computePriceIx = ComputeBudgetProgram.setComputeUnitPrice({
                    // @ts-ignore
                    microLamports: feeEstimate.priorityFeeEstimate,
                });
                const tx = txnsSigned[i];
                tx.add(computePriceIx);
                // resign
                const signers = ixAndSigners[i].signers;
                if (signers.length > 0) {
                    tx.partialSign(...signers);
                }
            }
        }
        console.log("Sending multiple transactions...");
        try {
            return await promiseAllInOrder(txnsSigned.map((txn, i) => async () => {
                console.log(`index ${i}`);
                const { txid } = await sendAndConfirmWithRetry(provider.connection, txn.serialize(), {
                    skipPreflight: true,
                }, finality, timeout);
                return txid;
            }));
        }
        catch (e) {
            console.error(e);
            const wrappedE = ProgramError.parse(e, idlErrors);
            throw wrappedE == null ? e : wrappedE;
        }
    };
    const recur = async (priorityLevel) => {
        try {
            return await sendMultipleTxs(priorityLevel);
        }
        catch (e) {
            // @ts-ignore
            let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
            let newPrio;
            if (priorityLevel === 'NONE') {
                newPrio = 'LOW';
            }
            else if (priorityLevel == 'LOW') {
                newPrio = 'MEDIUM';
            }
            else if (priorityLevel == 'MEDIUM') {
                newPrio = 'HIGH';
            }
            else if (priorityLevel == 'HIGH') {
                newPrio = 'VERY_HIGH';
            }
            const msg = `${baseMsg}. ${!!newPrio ? `retrying at ${newPrio}` : `hard failed`}`;
            console.log(msg);
            if (!newPrio) {
                throw e;
            }
            return await recur(newPrio);
        }
    };
    return recur(priorityLevel);
};
const SEND_TRANSACTION_INTERVAL = 10;
export const awaitTransactionSignatureConfirmation = async (txid, timeout, connection, commitment = "recent", queryStatus = false) => {
    let done = false;
    let status = {
        slot: 0,
        confirmations: 0,
        err: null,
    };
    let subId = 0;
    status = await new Promise(async (resolve, reject) => {
        const t = setTimeout(() => {
            if (done) {
                return;
            }
            done = true;
            console.log("Rejecting for timeout...");
            reject({ timeout: true });
        }, timeout);
        try {
            subId = connection.onSignature(txid, (result, context) => {
                status = {
                    err: result.err,
                    slot: context.slot,
                    confirmations: 0,
                };
                done = true;
                if (result.err) {
                    console.log(`Rejected ${txid} via websocket`, result.err);
                    reject(status);
                }
                else {
                    resolve(status);
                }
            }, commitment);
        }
        catch (e) {
            console.error("WS error in setup", txid, e);
        }
        finally {
            done = true;
            clearTimeout(t);
        }
        while (!done && queryStatus) {
            // eslint-disable-next-line no-loop-func
            (async () => {
                try {
                    const signatureStatuses = await connection.getSignatureStatuses([
                        txid,
                    ]);
                    status = signatureStatuses && signatureStatuses.value[0];
                    if (!done) {
                        if (!status) {
                        }
                        else if (status.err) {
                            console.log("REST error for", txid, status);
                            done = true;
                            reject(status.err);
                        }
                        else if (!status.confirmations && !status.confirmationStatus) {
                            console.log("REST no confirmations for", txid, status);
                        }
                        else {
                            console.log("REST confirmation for", txid, status);
                            if (!status.confirmationStatus ||
                                status.confirmationStatus == commitment) {
                                done = true;
                                resolve(status);
                            }
                        }
                    }
                }
                catch (e) {
                    if (!done) {
                        console.log("REST connection error: txid", txid, e);
                    }
                }
            })();
            await sleep(2000);
        }
    });
    if (
    //@ts-ignore
    connection._signatureSubscriptions &&
        //@ts-ignore
        connection._signatureSubscriptions[subId]) {
        connection.removeSignatureListener(subId);
    }
    done = true;
    return status;
};
async function simulateTransaction(connection, transaction, commitment) {
    // @ts-ignore
    transaction.recentBlockhash = await connection._recentBlockhash(
    // @ts-ignore
    connection._disableBlockhashCaching);
    const signData = transaction.serializeMessage();
    // @ts-ignore
    const wireTransaction = transaction._serialize(signData);
    const encodedTransaction = wireTransaction.toString("base64");
    const config = { encoding: "base64", commitment };
    const args = [encodedTransaction, config];
    // @ts-ignore
    const res = await connection._rpcRequest("simulateTransaction", args);
    if (res.error) {
        throw new Error("failed to simulate transaction: " + res.error.message);
    }
    return res.result;
}
const DEFAULT_TIMEOUT = 60; // 1 minute
/*
    A validator has up to 120s to accept the transaction and send it into a block.
    If it doesn’t happen within that timeframe, your transaction is dropped and you’ll need
    to send the transaction again. You can get the transaction signature and periodically
    Ping the network for that transaction signature. If you never get anything back,
    that means it’s definitely been dropped. If you do get a response back, you can keep pinging
    that means it’s definitely been dropped. If you do get a response back, you can keep pinging
    until it’s gone to a confirmed status to move on.
  */
export async function sendAndConfirmWithRetry(connection, txn, sendOptions, commitment, timeout = DEFAULT_TIMEOUT) {
    let done = false;
    let slot = 0;
    const txid = await connection.sendRawTransaction(txn, sendOptions);
    const startTime = getUnixTime();
    (async () => {
        while (!done && getUnixTime() - startTime < timeout) {
            console.log(`sending raw tx now - startTime ${getUnixTime() - startTime} < ${timeout}`);
            await connection.sendRawTransaction(txn, sendOptions);
            await sleep(500);
        }
    })();
    try {
        const confirmation = await awaitTransactionSignatureConfirmation(txid, timeout, connection, commitment, true);
        if (!confirmation)
            throw new Error("Timed out awaiting confirmation on transaction");
        if (confirmation.err) {
            const tx = await connection.getTransaction(txid, {
                commitment: "confirmed",
            });
            console.error(tx?.meta?.logMessages?.join("\n"));
            console.error(confirmation.err);
            throw new Error("Transaction failed: Custom instruction error");
        }
        slot = confirmation?.slot || 0;
    }
    catch (err) {
        console.error("Error caught", err);
        if (err.timeout) {
            throw new Error("Timed out awaiting confirmation on transaction");
        }
        const tx = await connection.getTransaction(txid, {
            commitment: "confirmed",
        });
        if (tx && tx.meta && tx.meta.logMessages) {
            console.error(tx.meta.logMessages.join("\n"));
        }
        if (err.err) {
            throw err.err;
        }
        throw err;
    }
    finally {
        done = true;
    }
    console.log("Latency", txid, getUnixTime() - startTime);
    return { txid };
}
export function stringToTransaction(solanaTransaction) {
    return Transaction.from(Buffer.from(solanaTransaction));
}
export function bufferToTransaction(solanaTransaction) {
    return Transaction.from(solanaTransaction);
}
async function withRetries(tries, input) {
    for (let i = 0; i < tries; i++) {
        try {
            return await input();
        }
        catch (e) {
            console.log(`Retrying ${i}...`, e);
        }
    }
    throw new Error("Failed after retries");
}
const TX_BATCH_SIZE = 200;
export async function bulkSendTransactions(provider, txs, onProgress, triesRemaining = 10 // Number of blockhashes to try resending txs with before giving up
) {
    let ret = [];
    // attempt to chunk by blockhash bounds (so signing doesn't take too long)
    for (let chunk of chunks(txs, TX_BATCH_SIZE)) {
        const thisRet = [];
        // Continually send in bulk while resetting blockhash until we send them all
        while (true) {
            const recentBlockhash = await withRetries(5, () => provider.connection.getLatestBlockhash("confirmed"));
            const signedTxs = await Promise.all(chunk.map(async (tx) => {
                tx.recentBlockhash = recentBlockhash.blockhash;
                // @ts-ignore
                const signed = await provider.wallet.signTransaction(tx);
                tx.signatures[0].signature;
                return signed;
            }));
            const txsWithSigs = signedTxs.map((tx, index) => {
                return {
                    transaction: chunk[index],
                    sig: bs58.encode(tx.signatures[0].signature),
                };
            });
            const confirmedTxs = await bulkSendRawTransactions(provider.connection, signedTxs.map((s) => s.serialize()), ({ totalProgress, ...rest }) => onProgress &&
                onProgress({ ...rest, totalProgress: totalProgress + ret.length + thisRet.length }), recentBlockhash.lastValidBlockHeight, 
            // Hail mary, try with preflight enabled. Sometimes this causes
            // errors that wouldn't otherwise happen
            triesRemaining != 1);
            thisRet.push(...confirmedTxs);
            if (confirmedTxs.length == signedTxs.length) {
                break;
            }
            const retSet = new Set(thisRet);
            chunk = txsWithSigs
                .filter(({ sig }) => !retSet.has(sig))
                .map(({ transaction }) => transaction);
            triesRemaining--;
            if (triesRemaining <= 0) {
                throw new Error(`Failed to submit all txs after blockhashes expired, ${signedTxs.length - confirmedTxs.length} remain`);
            }
        }
        ret.push(...thisRet);
    }
    return ret;
}
// Returns the list of succesfully sent txns
// NOTE: The return signatures are ordered by confirmation, not by order they are passed
// This list should be in order. Seom txns may fail
// due to blockhash exp
export async function bulkSendRawTransactions(connection, txs, onProgress, lastValidBlockHeight, skipPreflight = true) {
    const txBatchSize = TX_BATCH_SIZE;
    let totalProgress = 0;
    const ret = [];
    if (!lastValidBlockHeight) {
        const blockhash = await withRetries(5, () => connection.getLatestBlockhash("confirmed"));
        lastValidBlockHeight = blockhash.lastValidBlockHeight;
    }
    for (let chunk of chunks(txs, txBatchSize)) {
        let currentBatchProgress = 0;
        let pendingCount = chunk.length;
        let txids = [];
        let lastRetry = 0;
        while (pendingCount > 0) {
            if ((await withRetries(5, () => connection.getBlockHeight())) >
                lastValidBlockHeight) {
                return ret;
            }
            // only resend txs every 4s
            if (lastRetry < new Date().valueOf() - 4 * 1000) {
                lastRetry = new Date().valueOf();
                txids = [];
                for (const tx of chunk) {
                    const txid = await connection.sendRawTransaction(tx, {
                        skipPreflight,
                    });
                    txids.push(txid);
                }
            }
            const statuses = await getAllTxns(connection, txids);
            const completed = statuses.filter((status) => status !== null);
            totalProgress += completed.length;
            currentBatchProgress += completed.length;
            onProgress &&
                onProgress({
                    totalProgress: totalProgress,
                    currentBatchProgress: currentBatchProgress,
                    currentBatchSize: txBatchSize,
                });
            const failures = completed
                .map((status) => status !== null && status.meta?.err)
                .filter(truthy);
            if (failures.length > 0) {
                console.error(failures);
                throw new Error("Failed to run txs");
            }
            ret.push(...txids
                .map((txid, idx) => statuses[idx] == null ? null : txid)
                .filter(truthy));
            chunk = chunk.filter((_, index) => statuses[index] === null);
            txids = txids.filter((_, index) => statuses[index] === null);
            pendingCount -= completed.length;
            await sleep(1000); // Wait one seconds before querying again
        }
    }
    return ret;
}
const MAX_GET_SIGNATURE_STATUSES_QUERY_ITEMS = 200;
async function getAllTxns(connection, txids) {
    return (await Promise.all(chunks(txids, MAX_GET_SIGNATURE_STATUSES_QUERY_ITEMS).map((txids) => connection.getTransactions(txids, {
        maxSupportedTransactionVersion: 0,
        commitment: "confirmed",
    })))).flat();
}
class SendIxError extends Error {
    logs;
    constructor(originalError) {
        //The newlines don't actually show up correctly in chai's assertion error, but at least
        // we have all the information and can just replace '\n' with a newline manually to see
        // what's happening without having to change the code.
        const logs = originalError.logs?.join('\n') || "error had no logs";
        super(originalError.message + "\nlogs:\n" + logs);
        this.stack = originalError.stack;
        this.logs = logs;
    }
}
//# sourceMappingURL=transaction.js.map